package be.iminds.ilabt.jfed.ui.cli2;

import be.iminds.ilabt.jfed.ui.cli2.instruction.Instruction;
import be.iminds.ilabt.jfed.ui.cli2.instruction.RunInstruction;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.DefaultOmittingSerializationUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SequenceWriter;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.dataformat.yaml.YAMLFactory;
import javafx.embed.swing.JFXPanel;
import org.slf4j.LoggerFactory;

import java.io.*;
import java.lang.reflect.InvocationTargetException;
import java.util.Map;

/**
 * ExperimenterCli
 */
@SuppressWarnings("AccessStaticViaInstance")
public class ExperimenterCli {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ExperimenterCli.class);

    private static final String KEY_EXPIRATION_TIME = "expiration-time";
    private static final String KEY_REQUEST_RSPEC = "request-rspec";

    public static void main(String[] args) throws IOException, ClassNotFoundException {
        int retval = main_helper(args, System.out, System.err, System.in, new BaseCli(System.err), true);
        System.exit(retval);
    }

    public static int main_helper(String[] args,
                                  final PrintStream out, final PrintStream err, final InputStream in,
                                  BaseCli baseCli,
                                  boolean temporarilyRedirectSystemStreams) throws IOException, ClassNotFoundException {
        final PrintStream orig_out = System.out;
        final PrintStream orig_err = System.err;
        final InputStream orig_in = System.in;
        if (temporarilyRedirectSystemStreams) {
            System.setOut(out);
            System.setErr(err);
            System.setIn(in);
        }

        int retval = -1;
        try {
            retval = main_internal(args, out, err, in, baseCli);
        } catch (CliExitException e) {
            return e.getExitCode();
        } finally {
            if (temporarilyRedirectSystemStreams) {
                System.setOut(orig_out);
                System.setErr(orig_err);
                System.setIn(orig_in);
            }
        }
        return retval;
    }

    public static int main_internal(String[] args,
                                    final PrintStream out, final PrintStream err, final InputStream in,
                                    BaseCli baseCli) throws CliExitException {
        //initialize JavaFX-toolkit
        new JFXPanel();

        final ExperimenterCommandLineParser experimenterCommandLineParser;

        experimenterCommandLineParser = new ExperimenterCommandLineParser(args, out, err, in, baseCli);
        ExperimenterCommandLineParser.GenericCliArguments genCliArgs = experimenterCommandLineParser.getGenericCliArguments();

        if (args.length == 0 || genCliArgs.isHelp()) {
            experimenterCommandLineParser.generalHelp(err);
            return 0;
        }
        if (genCliArgs.isVersion()) {
            experimenterCommandLineParser.showVersion(out);
            return 0;
        }

        if (genCliArgs.isDebug())
            Context.setupDebugOutputToStdout();

        Instruction instruction = null;
        try {
            instruction = experimenterCommandLineParser.processInstruction();
        } catch (CliExitException e) {
            //outputs not yet OK, so show error on stdout
            err.println(""+e.getMessage());
            throw e;
        }

        if (genCliArgs.isPrintAction()) {
            ObjectMapper om = new ObjectMapper();
            Instruction defaultInstruction = null;
            try {
                defaultInstruction = instruction.getClass().getConstructor(/*noargs*/).newInstance();
            } catch (Exception e) {
                throw new CliExitException(Context.ExitReason.ERROR,
                        "Implementation exception for "+instruction.getClass()+" (default constructor missing?)", e);
            }
            Map defaultInstructionMap = om.convertValue(defaultInstruction, Map.class);
            defaultInstructionMap.remove("action"); //action should never be seen as a default
            Map prunedInstructions = DefaultOmittingSerializationUtil.omitDefault(
                    om.convertValue(instruction, Map.class),
                    defaultInstructionMap
            );

            YAMLFactory yf = new YAMLFactory();
            ObjectMapper mapper = new ObjectMapper(yf);
            try (SequenceWriter sw = mapper.writerWithDefaultPrettyPrinter()
                    .withFeatures(SerializationFeature.ORDER_MAP_ENTRIES_BY_KEYS)
                    .writeValues(out)) {
                sw.write(prunedInstructions);
            } catch (IOException e) {
                throw new CliExitException(Context.ExitReason.ERROR, "Error while converting config to yml: "+e.getMessage(), e);
            }
            return 0;
        }

        Context context = new Context(args, genCliArgs, instruction, out, err, in, baseCli);

        try {
            //setup the outputs according to the instruction
            context.setupDebugOutput();
            context.setupActionOutput();
        } catch (CliExitException e) {
            //outputs not yet OK, so show error on stdout
            err.println(""+e.getMessage());
            throw e;
        }

        try {
            //note: thanks to context.setupDebugOutput() we can use default logging with static loggers
            LOG.debug("Final Instruction: " + context.getInstruction());

            context.setupInjector();
            context.setupCallOutput(); //needs Logger  (so take injector into account here)

            InstructionExecutor instructionExecutor = new InstructionExecutor(context);
            instructionExecutor.run(context, instruction);
            return 0;
        } catch (CliExitException e) {
            assert e.getExitReason() != Context.ExitReason.SUCCESS;
            if (e.getExitReason() != Context.ExitReason.SUCCESS && e.getMessage() != null) {
                context.getActionLogger().errorPrintln(e.getMessage());
            }
            return e.getExitCode();
        } finally {
            context.closeOutputs();
        }
    }

//    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, String sshKey) {
//        assert userUrn != null;
//        addToUserSpecs(userSpecs, userUrn, Collections.singletonList(sshKey));
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, String sshKey) {
//        addToUserSpecs(userSpecs, userUrnString, Collections.singletonList(sshKey));
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, GeniUrn userUrn, Collection<String> sshKeys) {
//        assert userUrn != null;
//        addToUserSpecs(userSpecs, userUrn.toString(), sshKeys);
//    }
//    private static void addToUserSpecs(List<UserSpec> userSpecs, String userUrnString, Collection<String> sshKeys) {
//        if (sshKeys == null || sshKeys.isEmpty())
//            return;
//        for (UserSpec us : userSpecs) {
//            if (Objects.equals(us.getUrn(), userUrnString)) {
//                //TODO (low priority) make this work for SSH keys that differ only in comment, but not in actual key
//                Set<String> newSskKeys = new HashSet<>(sshKeys);
//                newSskKeys.removeAll(us.getSshKey());
//                if (!newSskKeys.isEmpty())
//                    us.getSshKey().addAll(newSskKeys);
//                return;
//            }
//        }
//
//        UserSpec newUs = new UserSpec(userUrnString, sshKeys);
//        userSpecs.add(newUs);
//    }
}
